﻿#include <iostream>
#include <vector>
#include <algorithm>

#include "randomc.h"
#include "mersenne.cpp"

using namespace std;

const double NEVER = 1000000000.0;

class single_server
{
public:
	// model inputs
	double mean_time_between_arrivals;
	double mean_time_to_process;

	// state variables
	unsigned int queue;      // queue length
	unsigned int server;     // 0=idle, 1=working

	bool off_2_on;           // returns true if server switched from "idle" to "working"

	// time structure
	double t_now;
	vector<double> times_of_next_events;   // 1st element for "arrival", 2nd element for "departure"

	unsigned int event_id;                 // 0 for "arrival", 1 for "departure"

	// statistics
	unsigned int number_of_departures;
	unsigned int max_customers;

	double update_throughput()
	{
		return static_cast<double> (number_of_departures) / t_now;
	}

	void update_max_customers()
	{
		unsigned int current_customers = queue + server;

		if (current_customers > max_customers)
			max_customers = current_customers;
	}

	single_server(double some_mean_time_between_arrivals, double some_mean_time_to_process, TRandomMersenne* rand_gen)
	{
		mean_time_between_arrivals = some_mean_time_between_arrivals;
		mean_time_to_process = some_mean_time_to_process;

		queue = 0;
		server = 0;
		off_2_on = 0;

		t_now = 0.0;
		times_of_next_events.resize(2);
		times_of_next_events[0] = (-1.0) * log(rand_gen->Random()) * mean_time_between_arrivals;
		times_of_next_events[1] = NEVER;

		number_of_departures = 0;
		max_customers = 0;
	}

	void find_time_of_next_event()
	{
		t_now = (*min_element(times_of_next_events.begin(), times_of_next_events.end()));

		for (unsigned int i = 0; i < times_of_next_events.size(); i++)
		{
			if (times_of_next_events[i] == t_now)
				event_id = i;
		}
	}

	void execute_arrival()
	{
		queue++;
	}

	void execute_departure()
	{
		server = 0;
	}

	void event_execution()
	{
		if (event_id == 0)
			execute_arrival();

		if (event_id == 1)
		{
			execute_departure();
			number_of_departures++;
		}

		update_max_customers();
	}

	void controller()
	{
		if (queue > 0 && server == 0)
		{
			queue--;
			server = 1;
			off_2_on = 1;
		}
	}

	void update_times_of_next_events(TRandomMersenne* arrival_times, TRandomMersenne* process_times)
	{
		if (event_id == 0)
			times_of_next_events[0] = t_now + ((-1.0) * log(arrival_times->Random()) * mean_time_between_arrivals); 

		if (off_2_on)
		{
			times_of_next_events[1] = t_now + ((-1.0) * log(process_times->Random()) * mean_time_to_process);
			off_2_on = 0;
		}

		if (event_id == 1 && server == 0)
			times_of_next_events[1] = NEVER;
	}
};

int main()
{
	int32 seed1 = 123;
	int32 seed2 = 456;

	TRandomMersenne arrival_process(seed1);
	TRandomMersenne service_process(seed2);

	single_server model(1.6, 1.0, &arrival_process);

	while (model.number_of_departures < 10000)
	{
		model.find_time_of_next_event();
		model.event_execution();
		model.controller();
		model.update_times_of_next_events(&arrival_process, &service_process);
	}

	cout << "Simulation terminated....." << endl<<endl;
	cout << "Statistics:" << endl<<endl;
	cout << "	Simulation time: " << model.t_now << endl;
	cout << "	Total serviced customers: " << model.number_of_departures << endl;
	cout << "	System throughput: " << model.update_throughput() << endl;
	cout << "	Max customers: " << model.max_customers<< endl<<endl;
	cout << "End of experiment......" << endl << endl;

	return 0;
}