#include <cmath>
#include <vector>
#include <iostream>
#include <fstream>
#include <algorithm>
#include "randomc.h"
#include "mersenne.cpp"

using namespace std;
const double NEVER=1000000000.0;

class stage
{
public:
	unsigned int I;          // input queue for machine M
	unsigned int M;          // machine state: 1=working, 2=idle
	unsigned int PA;         // queue, finished parts with kanbans attached
	unsigned int A;          // queue, free kanbans
	unsigned int D;          // queue, demands
	unsigned int DA;         // queue, demands with kanbans attached

	unsigned int M_state_transition;    // 12=from working to idle, 21=from idle to working
};

typedef vector<stage> n_stages;

class stats
{
public:
	double meanWIP;     // mean work in process: contents of input queues+contents of machines+contents of output buffers except the last one
	double meanI;       // mean finished goods inventory
	double meanB;       // mean length of backorders

	double integralWIP; // the integral of WIP over time
	double integralI;   // the integral of finished goods inventory over time
	double integralB;   // the the integral of backorders over time

	stats()
	{
		meanWIP=meanI=meanB=0.0;
		integralWIP=integralI=integralB=0.0;
	}

	void updateWIP(const n_stages &line, double t_now, double t_before)
	{
		double WIP=0.0;

		for(unsigned int i=0;i<line.size();i++)
		{
			WIP+=max(0.0,static_cast<double>(line[i].I) );

			if(line[i].M==1)
				WIP+=1.0;

			if(i!=line.size()-1)
				WIP+=line[i].PA;
		}

		integralWIP+=WIP*(t_now-t_before);
		meanWIP=integralWIP/t_now;
	}

	void updateI(double I, double t_now, double t_before)
	{
		integralI+=I*(t_now-t_before);
		meanI=integralI/t_now;
	}

	void updateB(double B, double t_now, double t_before)
	{
		integralB+=B*(t_now-t_before);
		meanB=integralB/t_now;
	}

	void update_all(const n_stages &line, double I, double B, double t_now, double t_before)
	{
		updateWIP(line, t_now, t_before);
		updateI(I, t_now, t_before);
		updateB(B, t_now, t_before);
	}
};

class GK 
{ public:
// model parameters
unsigned int n;                            // number of machines
double mean_time_between_arrivals;     
vector<double> mean_processing_times;

//system state variables
vector<stage> flow_shop;
unsigned int Dn;                           // backorders

// clock structure
double t_now;                              // current time
double t_before;                           // time of previous event
vector<double> times_of_next_events;       // [0]: arrival, [1]: production in 1st stage, etc etc.....
unsigned int event_id;                     // 0=arrival, 1=production in 1st stage, etc etc.....

// simulation statistics
stats queue_stats;

GK(TRandomMersenne* rand_gen)
{	// initialize model parameters from file
	ifstream input_file;     input_file.open("instance.txt",ios::in);

	input_file>>n;                            // set number of stages
	flow_shop.resize(n);
	mean_processing_times.resize(n);
	
	input_file>>mean_time_between_arrivals;   // set mean time between arrivals

	for(unsigned int i=0;i<n;i++)
	{
		input_file>>mean_processing_times[i]; // set mean processing times
		input_file>>flow_shop[i].PA;          // set initial stocks   
		input_file>>flow_shop[i].A;           // set number of free kanbans
	}
	
	input_file.close();

	// initialize state variables 
	for(unsigned int i=0;i<n;i++)
	{	
		flow_shop[i].M = 2;
		flow_shop[i].M_state_transition=0;
		flow_shop[i].D=0;
		flow_shop[i].DA=0;
		flow_shop[i].I=0;
	}

	Dn=0;
	
	// initialize the clock structure 
	times_of_next_events.resize(1+n);	

	// arrival event is active always
	times_of_next_events[0]=(-1.0)*log(rand_gen->Random())*mean_time_between_arrivals;
    // production events are inactive initially
	for (unsigned int i=1; i<1+n; i++)
		times_of_next_events[i]=NEVER;
	
	event_id=0;
	t_now=t_before=0.0;
}

void find_time_of_next_event() 
{ 
t_before=t_now;      

t_now=(*min_element(times_of_next_events.begin(),times_of_next_events.end()) );

for(unsigned int i=0;i<times_of_next_events.size();i++)
{
	if(times_of_next_events[i]==t_now)
		event_id=i; 	
}

queue_stats.update_all(flow_shop, flow_shop[n-1].PA, Dn, t_now, t_before);
}

void execute_arrival()
{
  Dn++;
  flow_shop[n-1].D++;
}

void execute_production()
{
	flow_shop[event_id-1].PA++;
	flow_shop[event_id-1].M=2;
	flow_shop[event_id-1].M_state_transition=12;
}

void event_execution()
{       
	if (event_id != 0)
		execute_production();
	else
		execute_arrival(); 
}

void controller()
{
	if(Dn!=0 && flow_shop[n-1].PA!=0)
	{
		Dn--;
		flow_shop[n-1].PA--;
		flow_shop[n-1].A++;
	}

	for(unsigned int i=n-1;i>0;i--)
	{
		if(flow_shop[i].A!=0 && flow_shop[i].D!=0)
		{
			flow_shop[i].A--;
			flow_shop[i].D--;
			flow_shop[i].DA++;
			flow_shop[i-1].D++;
		}

		if(flow_shop[i].DA!=0 && flow_shop[i-1].PA!=0)
		{
			flow_shop[i].DA--;
			flow_shop[i-1].PA--;
			flow_shop[i].I++;
			flow_shop[i-1].A++;
		}
	}

	if(flow_shop[0].A!=0 && flow_shop[0].D!=0)
	{
		flow_shop[0].A--;
		flow_shop[0].D--;
		flow_shop[0].I++;
	}

	for(unsigned int i=0;i<n;i++)
	{
		if(flow_shop[i].I!=0 && flow_shop[i].M==2)
		{
			flow_shop[i].I--;
			flow_shop[i].M=1;
			flow_shop[i].M_state_transition=21;
		}
	}
}

void update_times_of_next_events(TRandomMersenne* arrival_process, TRandomMersenne* production_processes[]) 
{
	for(unsigned int i=0;i<n;i++)
	{// if some machine was switched to "working" then time of next production is updated
		if(flow_shop[i].M_state_transition==21)
		{
			times_of_next_events[i+1] = t_now + ( (-1.0) * log(production_processes[i]->Random()) * mean_processing_times[i]);
			flow_shop[i].M_state_transition=0;
		}
		// if some machine was swotched to "idle" then next production event is inactive
		if(flow_shop[i].M_state_transition==12)
		{
			times_of_next_events[i+1]=NEVER;
			flow_shop[i].M_state_transition=0;
		}
	}

	if(event_id==0)
	{// if there was an arrival then time of next arrival is updated
		times_of_next_events[event_id] = t_now + ((-1.0)*log(arrival_process->Random())*mean_time_between_arrivals);
	}
}
};
