﻿#include <iostream>
#include <vector>
#include <algorithm>

#include "randomc.h"
#include "mersenne.cpp"

const double NEVER = 1000000000.0;
using namespace std;

class KS
{
public:
	//model parameters
	unsigned int K;
	unsigned int d;
	double mean_time_between_arrivals;
	double mean_time_to_produce;
	double mean_time_to_fail;
	double mean_time_to_repair;

	// state variables
	unsigned int machine;           // 0=down, 1=working, 2=idle
	unsigned int deterioration;     // 0=good as new, 1=....., d+1=failed
	unsigned int buffer;            // number of finished goods
	unsigned int queue;             // number of pending production authorizations
	bool off2on;                    // 1 if machine switched to working state, 0 otherwise

	// clock structure
	double t_now;                           // current simulation time
	double t_before;                        // time of previous simulation event
	vector<double> times_of_next_events;    // [0]=for arrival, [1]=for production, [2]=for failure, [3]=for repair
	unsigned int event_id;                  // 0=arrival, 1=production, 2=deterioration failure, 3=repair

	// statistics
	double total_production;                // total number of produced goods    
	double total_failures;                  // total number of "hard" failures
	double total_up_time;                   // total time where the machine was operational
	double integral_buffer;                 // the integral of the finished goods inventory over time
	double mean_buffer;                     // average inventory over time

	KS(unsigned int some_K, unsigned int some_d, double some_mean_time_between_arrivals, double some_mean_time_to_produce, double some_mean_time_to_fail, double some_mean_time_to_repair, TRandomMersenne* rand_gen)
	{
		K = some_K;
		d = some_d;
		mean_time_between_arrivals = some_mean_time_between_arrivals;
		mean_time_to_produce = some_mean_time_to_produce;
		mean_time_to_fail = some_mean_time_to_fail;
		mean_time_to_repair = some_mean_time_to_repair;

		machine = 2;
		deterioration = 0;
		buffer = K;
		queue = 0;
		off2on = 0;

		t_now = t_before = 0.0;
		times_of_next_events.resize(4);
		times_of_next_events[0] = (-1.0) * log(rand_gen->Random()) * mean_time_between_arrivals;
		for (unsigned int i = 1; i < 4; i++)
		{
			times_of_next_events[i] = NEVER;
		}

		total_production = 0.0;
		total_failures = 0.0;
		total_up_time = 0.0;
		integral_buffer = 0.0;
		mean_buffer = 0.0;
	}

	void find_time_of_next_event()
	{
		t_before = t_now;
		t_now = (*min_element(times_of_next_events.begin(), times_of_next_events.end()));
		for (unsigned int i = 0; i < times_of_next_events.size(); i++)
		{
			if (times_of_next_events[i] == t_now)
				event_id = i;
		}
		//update statitsics
		integral_buffer += buffer * (t_now - t_before);
		mean_buffer = integral_buffer / t_now;
		if (machine == 1)
			total_up_time += t_now - t_before;
	}

	void execute_arrival()
	{
		if (buffer >= 1)
		{
			buffer--;
			queue++;
		}
	}

	void execute_production()
	{
		buffer++;
		machine = 2;
		//update statistics
		total_production += 1.0;
	}

	void execute_failure()
	{
		if (deterioration <= d)
			deterioration++;

		if (deterioration == d + 1)
		{
			machine = 0;
			queue++;                   // production authorization is sent back to queue to resume production after repair
			// update statistics
			total_failures += 1.0;
		}
	}

	void execute_repair()
	{
		deterioration = 0;
		machine = 2;
	}

	void event_execution()
	{
		if (event_id == 0)
			execute_arrival();

		if (event_id == 1)
			execute_production();

		if (event_id == 2)
			execute_failure();

		if (event_id == 3)
			execute_repair();
	}

	void controller()
	{
		if (queue > 0 && machine == 2)
		{
			queue--;
			machine = 1;
			off2on = 1;
		}
	}

	void update_times_of_next_events(TRandomMersenne* arrival_process, TRandomMersenne* production_process, TRandomMersenne* failure_process, TRandomMersenne* repair_process)
	{
		// if there was an arrival, update time of next arrival
		if (event_id == 0)
			times_of_next_events[0] = t_now + ((-1.0) * log(arrival_process->Random()) * mean_time_between_arrivals);

		// if machine switched to working from idle, update times of next production and deterioration failure
		if (off2on)
		{
			times_of_next_events[1] = t_now + ((-1.0) * log(production_process->Random()) * mean_time_to_produce);
			times_of_next_events[2] = t_now + ((-1.0) * log(failure_process->Random()) * mean_time_to_fail);
			off2on = 0;
		}

		// if there was a production 
		if (event_id == 1)  
		{// and the machine started a new production cycle update the time of next production
			if (machine == 1)            
				times_of_next_events[1] = t_now + ((-1.0) * log(production_process->Random()) * mean_time_to_produce);
			else
			{// and the machine remained idle afterwards, the production and deterioration failure events are inactive
				times_of_next_events[1] = NEVER;
				times_of_next_events[2] = NEVER; 
			}
		}

		// if there was a deterioration failure 
		if (event_id == 2)
		{// and it was a "hard" failure then the next repair event is active and the next production and failure events are inactive
			if (deterioration == d + 1)        
			{
				times_of_next_events[1] = NEVER;
				times_of_next_events[2] = NEVER;
				times_of_next_events[3] = t_now + ((-1.0) * log(repair_process->Random()) * mean_time_to_repair);
			}// and it was a "soft" failure update only the time of next deterioration failure
			else             
				times_of_next_events[2] = t_now + ((-1.0) * log(failure_process->Random()) * mean_time_to_fail);
		}

		// if there was a repair event
		if (event_id == 3)
		{// then the repair becomes inactive
			times_of_next_events[3] = NEVER;  
			// if the machine was switched to working then the production and deterioration failure events become active
			if (machine == 1)                      
			{
				times_of_next_events[1] = t_now + ((-1.0) * log(production_process->Random()) * mean_time_to_produce);
				times_of_next_events[2] = t_now + ((-1.0) * log(failure_process->Random()) * mean_time_to_fail);
			}
		}
	}
};

int main()
{
	int32 seed1 = 123;
	int32 seed2 = 456;
	int32 seed3 = 789;
	int32 seed4 = 101112;

	TRandomMersenne arrival_process(seed1);
	TRandomMersenne production_process(seed2);
	TRandomMersenne failure_process(seed3);
	TRandomMersenne repair_process(seed4);

	//initialize parameter K, d, mean time between arrivals, mean time to produce, mean time to fail, mean time to repair 
	KS model(20, 2, 1.0, 1.0 / 2.0, 1 / 0.5, 1/0.7, &arrival_process);
	//set criterion for terminating simulation
	double max_production = 500000.0;

	while (model.total_production < max_production)
	{
		model.find_time_of_next_event();
		model.event_execution();
		model.controller();
		model.update_times_of_next_events(&arrival_process, &production_process, &failure_process, &repair_process);
	}

	cout << "Simulation terminated at time: " << model.t_now << endl;
	cout << "Statistics:" << endl;
	cout << "	Throughput: " << model.total_production / model.t_now << endl;
	cout << "	Average finished goods inventory: " << model.mean_buffer << endl;
	cout << "	Rate of occurence of hard failures: " << model.total_failures / model.t_now << endl;
	cout << "	Machine utilization: " << model.total_up_time / model.t_now << endl;

	return 0;
}